﻿using System;
using System.IO;
using System.Linq;
using System.Net;
using System.Security.Cryptography;
using System.Text;
using System.Windows.Forms;

namespace NAND_Extractor
{
    public partial class nandExtractor : Form
    {
        public nandExtractor()
        {
            InitializeComponent();
        }


        /*
         * Class for fst entries.
         */
        public class fst_t
        {
            public byte[] filename = new byte[0x0B];
            public byte mode;
            public byte attr;
            public UInt16 sub;
            public UInt16 sib;
            public UInt32 size;
            public UInt32 uid;
            public UInt16 gid;
            public UInt32 x3;
        }


        /*
         * Globals.
         */
        public static byte[] key = new byte[16];
        public static byte[] iv = new byte[16];
        public static BinaryReader rom;
        public static Int32 loc_super;
        public static Int32 loc_fat;
        public static Int32 loc_fst;
        public static string nandFilename;
        public static string extractPath;
        public static int nand_type = -1;


        /*
         * Core functions.
         */
        private void fileOpen()
        {
            

            FileDialog fd = new OpenFileDialog();
            fd.Filter = "NAND dumps|*.bin|All files|*.*";
            fd.Title = "Open NAND dump file";

            if (fd.ShowDialog(this) != DialogResult.Cancel)
            {
                string filename = fd.FileName.Remove(0, fd.FileName.LastIndexOf("\\") + 1);

                statusText(string.Format("Loading {0} for viewing...", filename));

                int i = filename.LastIndexOf(".");
                if (i > 0)
                    filename = filename.Remove( filename.LastIndexOf("."), 
                                                filename.Length - filename.LastIndexOf("."));

                extractPath = Directory.GetCurrentDirectory() + "\\" + filename;
                if (File.Exists(extractPath))
                    extractPath += "-extract";

                fileView.Nodes.Clear();
                fileView.Nodes.Add("0000", filename);

                nandFilename = fd.FileName;

                if(initNand())
                {
                    viewFST(0, 0);

                    fileView.Sort();
                    fileView.Nodes["0000"].Expand();
                }

                rom.Close();

                statusText(string.Empty);
            }
            
        }

        private bool initNand()
        {
            rom = new BinaryReader(File.Open(nandFilename, 
                                        FileMode.Open, 
                                        FileAccess.Read,
                                        FileShare.Read), 
                                    Encoding.ASCII);

                
            nand_type = getDumpType(rom.BaseStream.Length);

            if (getKey(nand_type))
            {
                    //Console.WriteLine(BitConverter.ToString(key).Replace("-", string.Empty));

                    try
                    {
                        loc_super = findSuperblock();
                    }
                    catch
                    {
                        statusText("Invalid or non-ECC NAND dump");
                        fileView.Nodes.Clear();
                        msg_Error("Can't find superblock.\nAre you sure this is a Full (with ECC) or BootMii NAND dump?");
                        return false;
                    }

                    //Console.WriteLine("Superblock @ {0}", loc_super);

                    loc_fat = loc_super;
                    loc_fst = loc_fat + 0x0C + 0x010800;
                    return true;
                
            }
            return false;
        }

        private int getDumpType(Int64 FileSize)
        {
            Int64[] modes = { 536870912,    // mode 0 | 536870912 == no ecc
                              553648128,    // mode 1 | 553648128 == ecc
                              553649152 };  // mode 2 | 553649152 == old bootmii

            for (int i = 0; i < modes.Count(); i++)
                if (modes[i] == FileSize)
                    return i;

            return -1;
        }

        private bool getKey(int mode)
        {
            switch (mode)
            {
                case 0:
                    msg_Error("I can't do non-ECC dumps yet, come back after the next release...");
                    return false;
                
                case 1:
                    string keyfile = Directory.GetCurrentDirectory() + "\\keys.bin";
                    if (File.Exists(keyfile))
                    {
                        BinaryReader keyReader = new BinaryReader(File.Open(keyfile,
                                    FileMode.Open,
                                    FileAccess.Read,
                                    FileShare.Read),
                                    Encoding.ASCII);
                        keyReader.BaseStream.Seek(0x158, SeekOrigin.Begin);
                        if (keyReader.Read(key, 0, 16) > 0)
                            return true;
                    }
                    break;
                
                case 2:
                    rom.BaseStream.Seek(0x21000158, SeekOrigin.Begin);
                    if (rom.Read(key, 0, 16) > 0)
                        return true;
                    break;
            }

            if (Properties.Settings.Default.nand_key.Length == 32)
            {
                key = strToByte(Properties.Settings.Default.nand_key);
                msg_Info(string.Format("No new key data found, using manually entered key\n{0}\n\nMAKE SURE THIS IS THE RIGHT KEY OR YOUR\nEXTRACTED FILES WILL NOT DECRYPT CORRECTLY!", 
                    BitConverter.ToString(key).Replace("-", string.Empty) ));
                return true;
            }
            msg_Error("Something went horribly wrong, I can't find the key.\nTry entering it manually from the File menu.");
            return false;
        }
        
        private Int32 findSuperblock()
        {
            Int32 loc, current, last = 0;

            int start   = 0x20BE0000;
            int end     = 0x21000000;
            int len     = 0x42000;

            rom.BaseStream.Seek(start + 4, SeekOrigin.Begin);

            for (loc = start; loc < end; loc += len)
            {
                current = (int) bswap(rom.ReadUInt32());
                
                if (current > last)
                    last = current;
                else
                    return loc - len;

                rom.BaseStream.Seek(len - 4, SeekOrigin.Current);
            }

            return -1;
        }

        private byte[] getCluster(UInt16 cluster_entry)
        {
            byte[] cluster = new byte[0x4000];
            byte[] page = new byte[0x840];

            rom.BaseStream.Seek(cluster_entry * 16896, SeekOrigin.Begin);

            for (int i = 0; i < 8; i++)
            {
                page = rom.ReadBytes(0x840);
                Buffer.BlockCopy(page, 0, cluster, i * 0x800, 0x800);
            }

            return aesDecrypt(key, iv, cluster);
        }

        private UInt16 getFAT(UInt16 fat_entry)
        {
            /*
             * compensate for "off-16" storage at beginning of superblock
             * 53 46 46 53   XX XX XX XX   00 00 00 00
             * S  F  F  S     "version"    padding?
             */
            fat_entry += 6;
            // location in fat of cluster chain
            int loc = loc_fat + ((((fat_entry / 0x400) * 0x20) + fat_entry) * 2);

            rom.BaseStream.Seek(loc, SeekOrigin.Begin);
            return rom.ReadUInt16();
        }


        /*
         * Viewer functions.
         */
        private void viewFST(UInt16 entry, UInt16 parent)
        {
            fst_t fst = new fst_t();
            
            // compensate for 64 bytes of ecc data every 64 fst entries
            int loc_entry = (((entry / 0x40) * 2) + entry) * 0x20;
            
            rom.BaseStream.Seek(loc_fst + loc_entry, SeekOrigin.Begin);

            fst.filename = rom.ReadBytes(0x0C);
            fst.mode = rom.ReadByte();
            fst.attr = rom.ReadByte();
            fst.sub = bswap(rom.ReadUInt16());
            fst.sib = bswap(rom.ReadUInt16());
            fst.size = bswap(rom.ReadUInt32());
            fst.uid = bswap(rom.ReadUInt32());
            fst.gid = bswap(rom.ReadUInt16());
            fst.x3 = bswap(rom.ReadUInt32());

            if (fst.sib != 0xffff)
                viewFST(fst.sib, parent);

            fst.mode &= 1;

            addEntry(fst, entry, parent);
        }

        private void addEntry(fst_t fst, UInt16 entry, UInt16 parent)
        {
            TreeNode[] node = fileView.Nodes.Find(parent.ToString("x4"), true);

            if (entry != 0)
            {
                if (node.Count() > 0)
                    node[0].Nodes.Add(entry.ToString("x4"), ASCIIEncoding.ASCII.GetString(fst.filename).
                                    Replace("\0", string.Empty));
                else
                    fileView.Nodes.Add(entry.ToString("x4"), ASCIIEncoding.ASCII.GetString(fst.filename).
                                    Replace("\0", string.Empty));
            }

            if (fst.mode == 0 && fst.sub != 0xffff)
                viewFST(fst.sub, entry);
        }
        

        /*
         * Extraction functions.
         */
        private void extractNAND()
        {
            statusText("Extracting NAND...");

            rom = new BinaryReader(File.Open(nandFilename,
                                                    FileMode.Open,
                                                    FileAccess.Read,
                                                    FileShare.Read),
                                                Encoding.ASCII);
            if (!Directory.Exists(extractPath))
                Directory.CreateDirectory(extractPath);

            extractFST(0, "");

            statusText(string.Empty);
            
            rom.Close();
        }

        private void extractFST(UInt16 entry, string parent)
        {
            fst_t fst = new fst_t();

            // compensate for 64 bytes of ecc data every 64 fst entries
            int loc_entry = (((entry / 0x40) * 2) + entry) * 0x20;

            rom.BaseStream.Seek(loc_fst + loc_entry, SeekOrigin.Begin);

            fst.filename = rom.ReadBytes(0x0C);
            fst.mode = rom.ReadByte();
            fst.attr = rom.ReadByte();
            fst.sub = bswap(rom.ReadUInt16());
            fst.sib = bswap(rom.ReadUInt16());
            fst.size = bswap(rom.ReadUInt32());
            fst.uid = bswap(rom.ReadUInt32());
            fst.gid = bswap(rom.ReadUInt16());
            fst.x3 = bswap(rom.ReadUInt32());

            if (fst.sib != 0xffff)
                extractFST(fst.sib, parent);

            fst.mode &= 1;

            switch (fst.mode)
            {
                case 0:
                    extractDir(fst, entry, parent);
                    break;
                case 1:
                    extractFile(fst, entry, parent);
                    break;
                default:
                    msg_Error(String.Format("Ignoring unsupported mode {0}.\n\t\t(FST entry #{1})",
                                                fst.mode, 
                                                entry.ToString("x4")));
                    break;
            }
        }

        private void extractSingleFST(UInt16 entry, string parent)
        {
            fst_t fst = new fst_t();
            
            if (!Directory.Exists(extractPath))
                Directory.CreateDirectory(extractPath);

            // compensate for 64 bytes of ecc data every 64 fst entries
            int loc_entry = (((entry / 0x40) * 2) + entry) * 0x20;

            rom.BaseStream.Seek(loc_fst + loc_entry, SeekOrigin.Begin);

            fst.filename = rom.ReadBytes(0x0C);
            fst.mode = rom.ReadByte();
            fst.attr = rom.ReadByte();
            fst.sub = bswap(rom.ReadUInt16());
            fst.sib = bswap(rom.ReadUInt16());
            fst.size = bswap(rom.ReadUInt32());
            fst.uid = bswap(rom.ReadUInt32());
            fst.gid = bswap(rom.ReadUInt16());
            fst.x3 = bswap(rom.ReadUInt32());

            fst.mode &= 1;

            switch (fst.mode)
            {
                case 0:
                    extractDir(fst, entry, parent);
                    break;
                case 1:
                    extractFile(fst, entry, parent);
                    break;
                default:
                    msg_Error(String.Format("Ignoring unsupported mode {0}.\n\t\t(FST entry #{1})",
                                                fst.mode, 
                                                entry.ToString("x4")));
                    break;
            }
        }

        private void extractDir(fst_t fst, UInt16 entry, string parent)
        {
            string filename = ASCIIEncoding.ASCII.GetString(fst.filename).Replace("\0", string.Empty);

            statusText(string.Format("Extracting:  {0}", filename));

            if (filename != "/")
            {
                if (parent != "/" && parent != "")
                    filename = parent + "\\" + filename;
             
                Directory.CreateDirectory(extractPath + "\\" + filename);
            }

            if (fst.sub != 0xffff)
                extractFST(fst.sub, filename);
        }

        private void extractFile(fst_t fst, UInt16 entry, string parent)
        {
            string filename = "\\" + parent + "\\" +
                            ASCIIEncoding.ASCII.GetString(fst.filename).Replace("\0", string.Empty);

            statusText(string.Format("Extracting:  {0}", filename));

            int cluster_span = (int) (fst.size / 0x4000) + 1;
            UInt16 fat;
            byte[] cluster = new byte[0x4000],
                   data = new byte[cluster_span * 0x4000];

            BinaryWriter bw = new BinaryWriter(File.Open(   extractPath + filename.Replace(":", "-"), 
                                                            FileMode.Create, 
                                                            FileAccess.Write, 
                                                            FileShare.Read ), 
                                                        Encoding.ASCII);

            fat = fst.sub;
            int i = 0;
            while (fat < 0xFFF0)
            {
                //Console.WriteLine("{0} @ cluster {1}", filename, fat);
                Buffer.BlockCopy(getCluster(fat), 0, data, i*0x4000, 0x4000);
                fat = bswap(getFAT(fat));
                i++;
            }

            bw.Write(data, 0, (int)fst.size);
            bw.Close();
        }



        /*
         * Crypto functions (encryption unused, but included for reference).
         * Key required length of 16 bytes.
         * IV can be from 1 to 16 byte(s) and will be padded with 0x00.
         */
        private byte[] aesDecrypt(byte[] key, byte[] iv, byte[] enc_data)
        {
            // zero out any remaining iv bytes
            byte[] iv16 = new byte[16];
            Buffer.BlockCopy(iv, 0, iv16, 0, iv.Length);

            RijndaelManaged aes = new RijndaelManaged();
            aes.Padding = PaddingMode.None;
            aes.Mode = CipherMode.CBC;

            ICryptoTransform decryptor = aes.CreateDecryptor(key, iv16);
            MemoryStream memoryStream = new MemoryStream(enc_data);
            CryptoStream cryptoStream = new CryptoStream(memoryStream,
                                                      decryptor,
                                                      CryptoStreamMode.Read);

            byte[] dec_data = new byte[enc_data.Length];

            int decryptedByteCount = cryptoStream.Read(dec_data, 0,
                                                   dec_data.Length);

            memoryStream.Close();
            cryptoStream.Close();

            //Console.WriteLine("Decrypted {0} bytes:", decryptedByteCount);

            Application.DoEvents();
            return dec_data;
        }

        private byte[] aesEncrypt(byte[] key, byte[] iv, byte[] dec_data)
        {
            // zero out any remaining iv bytes
            byte[] iv16 = new byte[16];
            Buffer.BlockCopy(iv, 0, iv16, 0, iv.Length);

            RijndaelManaged aes = new RijndaelManaged();
            aes.Padding = PaddingMode.None;
            aes.Mode = CipherMode.CBC;

            ICryptoTransform encryptor = aes.CreateEncryptor(key, iv16);
            MemoryStream memoryStream = new MemoryStream(dec_data);
            CryptoStream cryptoStream = new CryptoStream(memoryStream,
                                                      encryptor,
                                                      CryptoStreamMode.Read);

            byte[] enc_data = new byte[dec_data.Length];

            int encryptedByteCount = cryptoStream.Read(enc_data, 0,
                                                   enc_data.Length);

            memoryStream.Close();
            cryptoStream.Close();

            //Console.WriteLine("Encrypted {0} bytes:", encryptedByteCount);

            Application.DoEvents();
            return enc_data;
        }


        /*
         * Helper/misc functions.
         */
        public static void msg_Error(string message)
        {
            MessageBox.Show(Form.ActiveForm, message, "Error!", 
                                MessageBoxButtons.OK, 
                                MessageBoxIcon.Error);
        }

        public static void msg_Info(string message)
        {
            MessageBox.Show(Form.ActiveForm, message, "Information!", 
                                MessageBoxButtons.OK, 
                                MessageBoxIcon.Information);
        }

        public static UInt16 bswap(UInt16 value)
        {
            return (UInt16)((0x00FF & (value >> 8))
                             | (0xFF00 & (value << 8)));
        }

        public static UInt32 bswap(UInt32 value)
        {
            UInt32 swapped = ((0x000000FF) & (value >> 24)
                             | (0x0000FF00) & (value >> 8)
                             | (0x00FF0000) & (value << 8)
                             | (0xFF000000) & (value << 24));
            return swapped;
        }

        public static byte[] strToByte(string hexString)
        {
            hexString = System.Text.RegularExpressions.Regex.Replace(hexString.ToUpper(), "[^0-9A-F]", string.Empty);
            byte[] b = new byte[hexString.Length / 2];
            for (int i = 0; i < hexString.Length; i += 2)
                b[i / 2] = byte.Parse(hexString.Substring(i, 2), System.Globalization.NumberStyles.AllowHexSpecifier);

            return b;
        }

        public void statusText(string message)
        {
            status.Text = message.Replace("\\", "/");
            if (message == string.Empty)
            {
                contextExtract.Enabled = true;
                extractToolStripMenuItem.Enabled = true;
            }
            else
            {
                contextExtract.Enabled = false;
                extractToolStripMenuItem.Enabled = false;
            }
            Application.DoEvents();
        }

        public void nandExtractor_Resize(object sender, System.EventArgs e)
        {
            this.fileView.Width = Size.Width - 28;
            this.fileView.Height = Size.Height - 80;
        }


        /*
         * Menu functions.
         */
        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            fileOpen();
        }

        private void extractToolStripMenuItem_Click(object sender, EventArgs e)
        {
            var stopwatch = System.Diagnostics.Stopwatch.StartNew();

            extractNAND();

            stopwatch.Stop();

            msg_Info(string.Format("Finished extraction in {0}.", stopwatch.Elapsed));
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            MessageBox.Show(this,   "Wii NAND Extractor\n" +
                                    "Version " + NAND_Extractor.Program.version + "\n\n" +
                                    "(c) Ben Wilson / parannoyed\n" +
                                    "http://sites.google.com/site/parannoyedwii/", "About",
                                MessageBoxButtons.OK,
                                MessageBoxIcon.Question);
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void contextExtract_Click(object sender, EventArgs e)
        {
            if (fileView.SelectedNode == null)
                msg_Error("Try choosing a file/directory.");
            else
            {
                var stopwatch = System.Diagnostics.Stopwatch.StartNew();

                rom = new BinaryReader(File.Open(nandFilename,
                                                    FileMode.Open,
                                                    FileAccess.Read,
                                                    FileShare.Read),
                                                Encoding.ASCII);
                if (rom.BaseStream.Length > 0)
                    extractSingleFST(Convert.ToUInt16(fileView.SelectedNode.Name, 16), "");

                rom.Close();

                stopwatch.Stop();

                statusText(string.Empty);
                msg_Info( string.Format("Finished extraction in {0}.", stopwatch.Elapsed) );
            }
        }

        private void fileView_MouseDown(object sender, MouseEventArgs e)
        {
            if(e.Button != MouseButtons.Right)
                return;
            fileView.SelectedNode = fileView.GetNodeAt(e.X,e.Y);
        }

        private void enterNandKeyMenuItem_Click(object sender, EventArgs e)
        {
            Form frm = new NandKey();
            frm.ShowDialog(this);
        }
    }
}
